/* This file is part of the project "Hilbert II" - http://www.qedeq.org
 *
 * Copyright 2000-2014,  Michael Meyling <mime@qedeq.org>.
 *
 * "Hilbert II" is free software; you can redistribute
 * it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */

package org.qedeq.kernel.bo.service.unicode;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Locale;

import org.qedeq.base.io.Parameters;
import org.qedeq.base.io.TextOutput;
import org.qedeq.base.trace.Trace;
import org.qedeq.kernel.bo.KernelContext;
import org.qedeq.kernel.bo.log.QedeqLog;
import org.qedeq.kernel.bo.module.InternalModuleServiceCall;
import org.qedeq.kernel.bo.module.InternalServiceJob;
import org.qedeq.kernel.bo.module.KernelQedeqBo;
import org.qedeq.kernel.bo.service.basis.ModuleServicePluginExecutor;
import org.qedeq.kernel.se.common.ModuleService;
import org.qedeq.kernel.se.common.SourceFileExceptionList;


/**
 * Transfer a QEDEQ module into a UTF-8 text file.
 * <p>
 * <b>This is just a quick written generator. This class just generates some text output to be able
 * to get a visual impression of a QEDEQ module.</b>
 *
 * @author  Michael Meyling
 */
public class Qedeq2Utf8Executor implements ModuleServicePluginExecutor {

    /** This class. */
    private static final Class CLASS = Qedeq2Utf8Executor.class;

    /** Output goes here. */
    private TextOutput printer;

    /** Current destination file. */
    private File destination;

    /** Visit all nodes with this visitor. */
    private final Qedeq2UnicodeVisitor visitor;

    /** Break automatically before this column number. */
    private int maxColumns;

    /** Generate text for these languages. */
    private String[] languages;

    /** Current language selection. See {@link #languages}. */
    private int run = 0;

    /**
     * Constructor.
     *
     * @param   plugin      This plugin we work for.
     * @param   prop        QEDEQ BO object.
     * @param   parameters  Plugin parameter.
     */
    public Qedeq2Utf8Executor(final ModuleService plugin, final KernelQedeqBo prop,
            final Parameters parameters) {
        final boolean info = parameters.getBoolean("info");
        // automatically line break after this column. 0 means no automatic line breaking
        maxColumns = parameters.getInt("maximumColumn");
        if (parameters.getInt("maximumColumn") != 0) {
            maxColumns = Math.max(10, maxColumns);
        }
        final boolean brief = parameters.getBoolean("brief");
        visitor = new Qedeq2UnicodeVisitor(plugin, prop, info , maxColumns, true, brief);
    }

    public Object executePlugin(final InternalModuleServiceCall call, final Object data) {
        final String method = "executePlugin()";
        try {
            QedeqLog.getInstance().logRequest("Generate UTF-8", visitor.getKernelQedeqBo().getUrl());
            languages = visitor.getKernelQedeqBo().getSupportedLanguages();
            for (run = 0; run < languages.length; run++) {
                final String result = generateUtf8(call.getInternalServiceProcess(), languages[run], "1");
                if (languages[run] != null) {
                    QedeqLog.getInstance().logSuccessfulReply(
                        "UTF-8 for language \"" + languages[run]
                        + "\" was generated into \"" + result + "\"", visitor.getKernelQedeqBo().getUrl());
                } else {
                    QedeqLog.getInstance().logSuccessfulReply(
                        "UTF-8 for default language "
                        + "was generated into \"" + result + "\"", visitor.getKernelQedeqBo().getUrl());
                }
            }
            if (languages.length == 0) {
                QedeqLog.getInstance().logMessage("no supported language found, assuming 'en'");
                final String result = generateUtf8(call.getInternalServiceProcess(), "en", "1");
                QedeqLog.getInstance().logSuccessfulReply(
                    "UTF-8 for language \"en"
                    + "\" was generated into \"" + result + "\"", visitor.getKernelQedeqBo().getUrl());
            }
        } catch (final SourceFileExceptionList e) {
            final String msg = "Generation failed";
            Trace.fatal(CLASS, this, method, msg, e);
            QedeqLog.getInstance().logFailureReply(msg, visitor.getKernelQedeqBo().getUrl(), e.getMessage());
        } catch (IOException e) {
            final String msg = "Generation failed";
            Trace.fatal(CLASS, this, method, msg, e);
            QedeqLog.getInstance().logFailureReply(msg, visitor.getKernelQedeqBo().getUrl(), e.getMessage());
        } catch (final RuntimeException e) {
            Trace.fatal(CLASS, this, method, "unexpected problem", e);
            QedeqLog.getInstance().logFailureReply(
                "Generation failed", visitor.getKernelQedeqBo().getUrl(), "unexpected problem: "
                + (e.getMessage() != null ? e.getMessage() : e.toString()));
        }
        return null;
    }

    /**
     * Gives a UTF-8 representation of given QEDEQ module as InputStream.
     *
     * @param   process     This process executes us.
     * @param   language    Filter text to get and produce text in this language only. Might
     *                      be <code>null</code>
     * @param   level       Filter for this detail level. LATER mime 20050205: not supported
     *                      yet.
     * @return  Name of generated file.
     * @throws  SourceFileExceptionList Major problem occurred.
     * @throws  IOException     File IO failed.
     */
    public String generateUtf8(final InternalServiceJob process, final String language, final String level)
            throws SourceFileExceptionList, IOException {

        // first we try to get more information about required modules and their predicates..
        try {
            visitor.getKernelQedeqBo().getKernelServices().loadRequiredModules(
                process, visitor.getKernelQedeqBo());
            visitor.getKernelQedeqBo().getKernelServices().checkWellFormedness(
                process, visitor.getKernelQedeqBo());
        } catch (Exception e) {
            // we continue and ignore external predicates
            Trace.trace(CLASS, "generateUtf8(KernelQedeqBo, String, String)", e);
        }
        String lan = "en";
        if (language != null) {
            lan = language;
        }
//        if (level == null) {
//            this.level = "1";
//        } else {
//            this.level = level;
//        }
        String txt = visitor.getKernelQedeqBo().getModuleAddress().getFileName();
        if (txt.toLowerCase(Locale.US).endsWith(".xml")) {
            txt = txt.substring(0, txt.length() - 4);
        }
        if (lan.length() > 0) {
            txt = txt + "_" + lan;
        }
        destination = new File(KernelContext.getInstance().getConfig()
            .getGenerationDirectory(), txt + ".txt").getCanonicalFile();
        printer = new TextOutput(visitor.getKernelQedeqBo().getName(), new FileOutputStream(destination),
            "UTF-8");

        try {
            visitor.generateUtf8(process, printer, lan, level);
        } finally {
            printer.flush();
            printer.close();
        }
        if (printer.checkError()) {
            throw printer.getError();
        }
        return destination.toString();
    }

    public String getLocationDescription() {
        if (languages != null && run < languages.length) {
            return languages[run] + " " + visitor.getLocationDescription();
        }
        if (languages != null && languages.length > 0) {
            return languages[languages.length] + " " + visitor.getLocationDescription();
        }
        return "unknown";
    }

    public double getVisitPercentage() {
        return visitor.getVisitPercentage() / languages.length * (run + 1);
    }

    public boolean getInterrupted() {
        return visitor.getInterrupted();
    }

}
